import logging
import os
import subprocess
import sys

from art import text2art
from rich.console import Console
from rich.logging import RichHandler
from rich.text import Text
from rich.traceback import install
from rich_gradient import Gradient

install(show_locals=True)
console = Console()

logging.basicConfig(
    level="NOTSET",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[
        RichHandler(rich_tracebacks=True, markup=True, show_path=False, show_time=False)
    ],
)


def show_logo(text, font="small", style="bold", ptb=1):
    logo_art = text2art(text, font=font)
    if isinstance(logo_art, str):
        lines = str(logo_art).splitlines()
        artlol = Gradient(lines)
        console.print(artlol, style=style)
        console.line(ptb)


def run_commands(commands, quietly, tasker: bool = False):
    """
    Run commands with support for conditional execution based on directory existence.

    Args:
        commands: List of commands or list of command dictionaries
        quietly: Run all commands quietly unless overridden per command
        tasker: If True, disables progress messages
    """

    def run(cmd, quiet_mode, title: str = ""):
        try:
            if quiet_mode:
                if not tasker and title:
                    msg.progress(title)
                subprocess.run(
                    cmd,
                    shell=True,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    env=os.environ,
                )
            else:
                subprocess.run(cmd, shell=True, check=True, env=os.environ)
        except subprocess.CalledProcessError as e:
            if e.returncode == 130:
                msg.warning("Execution cancelled by user (Ctrl+C).")
                sys.exit(2)
            else:
                msg.warning(f"Command failed: {cmd}")
                msg.error(e)
                sys.exit(1)
        except KeyboardInterrupt:
            msg.warning("Execution cancelled by user.")
            sys.exit(2)  # Custom exit code for cancel
        except Exception as e:
            msg.error(f"Unexpected error running command: {cmd}")
            msg.error(e)
            sys.exit(1)

    if isinstance(commands, list):
        for command in commands:
            if isinstance(command, str):
                run(command, quietly)
            elif isinstance(command, dict):
                cmd = command.get("run")
                title = command.get("title", "")
                quiet = command.get("quiet", quietly)
                if cmd:
                    run(cmd, quiet, title)


class MessagePrinter:
    def print(self, message, **kwargs):
        color = kwargs.pop("color", None)
        bold = kwargs.pop("bold", False)
        inline = kwargs.pop("inline", False)
        prefix = kwargs.pop("prefix", None)
        inlast = kwargs.pop("inlast", False)
        styled_message = Text()
        if prefix:
            styled_message.append(f"{prefix} ", style="bold")

        style_str = f"bold {color}" if bold and color else color or ""
        styled_message.append(Text.from_markup(message, style=style_str))

        if inline:
            console.print(
                styled_message, end=" ", soft_wrap=True, highlight=True, markup=True
            )
            if inlast:
                console.print(" " * 5)
        else:
            console.print(
                styled_message,
                soft_wrap=True,
                markup=True,
                justify="left",
                highlight=True,
            )

    def success(self, message, **kwargs):
        kwargs.setdefault("color", "green")
        kwargs.setdefault("prefix", "[*]")
        self.print(message, **kwargs)

    def warning(self, message, **kwargs):
        kwargs.setdefault("color", "yellow")
        kwargs.setdefault("prefix", "[~]")
        self.print(message, **kwargs)

    def error(self, message, **kwargs):
        kwargs.setdefault("color", "red")
        kwargs.setdefault("prefix", "[x]")
        self.print(message, **kwargs)

    def info(self, message, **kwargs):
        kwargs.setdefault("color", "cyan")
        kwargs.setdefault("prefix", "[!]")
        self.print(message, **kwargs)

    def progress(self, message, **kwargs):
        kwargs.setdefault("color", "magenta")
        kwargs.setdefault("prefix", "[$]")
        self.print(message, **kwargs)


msg = MessagePrinter()
log = logging.getLogger("demodapk")



import json
import os
import sys

import inquirer

import demodapk
from demodapk.utils import console

SCHEMA_PATH = os.path.join(os.path.dirname(demodapk.__file__), "schema.json")
SCHEMA_URL = (
    "https://raw.githubusercontent.com/Veha0001/DemodAPK/refs/heads/main/demodapk"
    "/schema.json"
)
SCHEMA_NETLIFY = "https://demodapk.netlify.app/schema.json"
CONFIG_FILE = "config.json"


def ensure_config(schema_value):
    """Open or create config.json and set $schema at the top."""
    config = {}

    if os.path.exists(CONFIG_FILE):
        with open(CONFIG_FILE, "r", encoding="utf-8") as f:
            try:
                config = json.load(f)
            except json.JSONDecodeError:
                console.log("config.json exists but is invalid JSON. Rewriting it.")

    # Insert $schema at the top by creating a new dict
    new_config = {"$schema": schema_value}
    for k, v in config.items():
        if k != "$schema":  # Avoid duplicates
            new_config[k] = v

    with open(CONFIG_FILE, "w", encoding="utf-8") as f:
        json.dump(new_config, f, indent=4)
        console.print(schema_value)
    console.log("Add selected $schema to ./config.json")


def get_schema():
    questions = [
        inquirer.List(
            "schema_index",
            message="Select a way of JSON Schema",
            choices=["pack", "netlify", "githubusercontent"],
            default="netlify",
        )
    ]

    ans = inquirer.prompt(questions)
    choice = ans.get("schema_index") if ans else None

    if choice:
        console.log(f"[bold green]You selected Schema {choice}:[/bold green]")
    else:
        console.print("[red]No selection made[/red]")
        sys.exit(1)

    if choice == "pack":
        ensure_config(SCHEMA_PATH)
    elif choice == "githubusercontent":
        ensure_config(SCHEMA_URL)
    else:
        ensure_config(SCHEMA_NETLIFY)

    sys.exit(0)



import os
import re
import shutil
import sys
from contextlib import nullcontext
from os.path import abspath, basename

from platformdirs import user_config_path

from demodapk.baseconf import Apkeditor
from demodapk.tool import download_apkeditor, get_latest_version
from demodapk.utils import console, log, msg, run_commands


def update_apkeditor():
    """
    Ensure the latest APKEditor jar is present in the user config folder.
    Deletes older versions and downloads the latest.
    Returns the path to the latest jar.
    """
    config_dir = user_config_path("demodapk")
    os.makedirs(config_dir, exist_ok=True)

    latest_version = get_latest_version()
    if not latest_version:
        log.error("Could not fetch the latest APKEditor version.")
        return None

    # Remove all existing APKEditor jars
    for fname in os.listdir(config_dir):
        if re.match(r"APKEditor-(\d+\.\d+\.\d+)\.jar$", fname):
            path = os.path.join(config_dir, fname)
            try:
                os.remove(path)
                log.warning("Deleted: %s", path)
            except (PermissionError, shutil.Error):
                pass

    download_apkeditor(config_dir)
    latest_jar = os.path.join(config_dir, f"APKEditor-{latest_version}.jar")

    if os.path.exists(latest_jar):
        return latest_jar

    log.error("Failed to download APKEditor.")
    return None


def get_apkeditor_cmd(cfg: Apkeditor):
    """
    Return the command to run APKEditor.
    - Use the provided jar or pick the latest jar from config.
    - If missing, download the latest jar and prompt to rerun.
    """
    editor_jar = cfg.editor_jar
    javaopts = cfg.javaopts

    config_dir = user_config_path("demodapk")
    os.makedirs(config_dir, exist_ok=True)

    if editor_jar:
        if not os.path.exists(editor_jar):
            msg.error(f"Specified editor jar does not exist: {editor_jar}")
            sys.exit(1)
    else:
        jars = []
        for fname in os.listdir(config_dir):
            match = re.match(r"APKEditor-(\d+\.\d+\.\d+)\.jar$", fname)
            if match:
                version = tuple(int(x) for x in match.group(1).split("."))
                jars.append((version, os.path.join(config_dir, fname)))
        if jars:
            jars.sort(reverse=True)
            editor_jar = jars[0][1]

    if os.path.getsize(editor_jar) == 0:
        log.error("The APKEditor JAR is faulty.")
        update_apkeditor()
        sys.exit(0)
    # If jar  doesn't exist, update/download latest
    if not editor_jar or not os.path.exists(editor_jar):
        update_apkeditor()
        sys.exit(0)

    return f"java {javaopts} -jar {editor_jar}".strip()


def apkeditor_merge(
    cfg: Apkeditor,
    apk_file: str,
    merge_base_apk: str,
    quietly: bool,
    force: bool = False,
):
    # New base name of apk_file end with .apk
    command = f'{get_apkeditor_cmd(cfg)} m -i "{os.path.abspath(apk_file)}" -o "{os.path.abspath(merge_base_apk)}"'
    if force:
        command += " -f"
    msg.info(f"Merging: {os.path.basename(apk_file)}", bold=True, prefix="[-]")
    with (
        console.status(
            "[bold blue]Processing...", spinner="point", spinner_style="blue"
        )
        if quietly
        else nullcontext()
    ):
        run_commands([command], quietly, tasker=True)
    msg.info(
        f"Merged into: {os.path.basename(merge_base_apk)}",
        color="green",
        bold=True,
        prefix="[+]",
    )


def apkeditor_decode(
    cfg: Apkeditor,
    apk_file: str,
    output_dir: str,
    quietly: bool,
    force: bool,
):
    output_dir = abspath(output_dir)
    merge_base_apk = abspath(os.path.splitext(apk_file)[0] + ".apk")
    # If apk_file is not end with .apk then merge
    if apk_file.lower() != ".apk":
        if not os.path.exists(merge_base_apk):
            apkeditor_merge(cfg, apk_file, merge_base_apk, quietly)
        command = f'{get_apkeditor_cmd(cfg)} d -i "{merge_base_apk}" -o "{output_dir}"'
        apk_file = merge_base_apk
    else:
        command = f'{get_apkeditor_cmd(cfg)} d -i "{apk_file}" -o "{output_dir}"'

    if cfg.dex_option:
        command += " -dex"
    if force:
        command += " -f"
    msg.info(
        f"Decoding: [magenta underline]{basename(apk_file)}",
        bold=True,
        prefix="[-]",
    )
    with (
        console.status("[bold green]Processing...", spinner="point")
        if quietly
        else nullcontext()
    ):
        run_commands([command], quietly, tasker=True)
    msg.info(
        f"Decoded into: {cfg.to_output}",
        color="green",
        bold=True,
        prefix="[+]",
    )


def apkeditor_build(
    cfg: Apkeditor,
    input_dir: str,
    output_apk: str,
    quietly: bool,
    force: bool,
):
    input_dir = abspath(input_dir)
    output_apk = abspath(output_apk)
    command = f'{get_apkeditor_cmd(cfg)} b -i "{input_dir}" -o "{output_apk}"'
    if force:
        command += " -f"
    msg.info(f"Building: {basename(input_dir)}", bold=True, prefix="[-]")
    with (
        console.status("[bold green]Finishing Build...", spinner="point")
        if quietly
        else nullcontext()
    ):
        run_commands([command], quietly, tasker=True)
    if cfg.clean:
        output_apk = cleanup_apk_build(input_dir, output_apk)
    msg.info(
        f"Built into: {basename(output_apk)}",
        color="green",
        bold=True,
        prefix="[+]",
    )
    return output_apk


def cleanup_apk_build(input_dir: str, output_apk: str):
    dest_file = input_dir + ".apk"
    shutil.move(output_apk, dest_file)
    msg.info(f"Clean: {basename(input_dir)}")
    shutil.rmtree(input_dir, ignore_errors=True)
    return dest_file



"""
- https://github.com/textualize/rich/blob/master/examples/downloader.py
"""

import json
import os
import signal
import sys
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from threading import Event
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from demodapk.utils import log

# === Rich progress setup ===
progress = Progress(
    TextColumn(
        "[bold blue]{task.fields[filename]}",
        justify="right",
    ),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
)

done_event = Event()


def handle_sigint(signum, frame):
    done_event.set()
    _ = signum, frame


signal.signal(signal.SIGINT, handle_sigint)


# === File download function ===
def copy_url(task_id: TaskID, url: str, path: str) -> None:
    """Copy data from a url to a local file."""
    log.info("Requesting: %s", url)
    req = Request(url, headers={"User-Agent": "Mozilla/5.0"})
    with urlopen(req) as response:
        # Break if content length is missing
        total = response.info().get("Content-Length")
        total = int(total) if total is not None else 0

        progress.update(task_id, total=total)
        with open(path, "wb") as dest_file:
            progress.start_task(task_id)
            for data in iter(partial(response.read, 32768), b""):
                dest_file.write(data)
                progress.update(task_id, advance=len(data))
                if done_event.is_set():
                    return
    log.info("Downloaded: %s", path)


def download(urls, dest_dir="."):
    """Download multiple files to the given directory."""
    os.makedirs(dest_dir, exist_ok=True)
    with progress:
        with ThreadPoolExecutor(max_workers=4) as pool:
            for url in urls:
                filename = url.split("/")[-1]
                dest_path = os.path.join(dest_dir, filename)
                task_id = progress.add_task("download", filename=filename, start=False)
                pool.submit(copy_url, task_id, url, dest_path)


def get_latest_version():
    """Get the latest version of APKEditor from GitHub API."""
    url = "https://api.github.com/repos/reandroid/apkeditor/releases/latest"
    req = Request(url, headers={"User-Agent": "Mozilla/5.0"})
    try:
        with urlopen(req) as resp:
            data = json.load(resp)
            tag_name = data.get("tag_name")
            if tag_name:
                # Remove leading 'V' if present
                return tag_name.lstrip("Vv")
    except (URLError, HTTPError) as e:
        log.error(e)
        sys.exit(1)
    return None


def download_apkeditor(dest_path):
    latest_version = get_latest_version()
    if latest_version:
        log.info("APKEditor latest version: %s", latest_version)
        jar_url = (
            "https://github.com/REAndroid/APKEditor/releases/download/"
            f"V{latest_version}/APKEditor-{latest_version}.jar"
        )
        download([jar_url], dest_path)
    else:
        log.critical("Could not determine the latest version.")



"""DemodAPK CLI"""

from types import SimpleNamespace

import rich_click as click

from demodapk import __version__
from demodapk.baseconf import load_config
from demodapk.mods import dowhat, runsteps
from demodapk.utils import show_logo


@click.command()
@click.help_option("-h", "--help")
@click.argument(
    "apk_dir",
    required=False,
    type=click.Path(exists=True, file_okay=True, dir_okay=True, path_type=str),
    metavar="<apk>",
)
@click.option(
    "-c",
    "--config",
    default="config.json",
    type=click.Path(exists=False, file_okay=True, dir_okay=True, path_type=str),
    metavar="<json>",
    show_default=True,
    help="Path to the configuration file.",
)
@click.option(
    "-sc",
    "--schema",
    is_flag=True,
    help="Apply schema to the config.",
)
@click.option(
    "-S",
    "--single-apk",
    is_flag=True,
    default=False,
    help="Keep only the rebuilt APK.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force to overwrite.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(exists=False, file_okay=True, dir_okay=True, path_type=str),
    metavar="<path>",
    help="Path to writes decode and build.",
)
@click.option(
    "-ua",
    "--update-apkeditor",
    is_flag=True,
    help="Update APKEditor latest version.",
)
@click.option(
    "-dex",
    "--raw-dex",
    is_flag=True,
    default=False,
    help="Decode with raw dex.",
)
@click.option(
    "-n",
    "--no-rename-package",
    is_flag=True,
    help="Skip rename package.",
)
@click.option(
    "-nfb",
    "--no-facebook",
    is_flag=True,
    help="Skip Facebook API update.",
)
@click.option(
    "-nas",
    "--rename-smali",
    is_flag=True,
    help="Rename package in smali files and directories.",
)
@click.version_option(
    __version__,
    "-v",
    "--version",
)
def main(**kwargs):
    """DemodAPK: APK Modification Script"""
    # kwargs will contain all arguments/options
    args = SimpleNamespace(**kwargs)
    packer = load_config(args.config).get("DemodAPK", {})
    show_logo("DemodAPK")
    dowhat(args, click)
    runsteps(args, packer)


if __name__ == "__main__":
    main()
